<?php

/**
 * Copyright (c) 2024 Zend by Perforce
 */

namespace ZRay;

use JsonException;
use ZRayPlugin;

class CodeIgniter4
{
    private ?ZRayPlugin $zrp = null;
    private array $collected = [
        'Events' => [],
        'Listeners' => [],
        'Helpers' => [],
        'State' => [],
        'Routes' => [],
        'Response' => []
    ];

    /**
     * Create ZRayPlugin instance and set up first trace function
     */
    public function __construct()
    {
        $this->zrp = new ZRayPlugin("CodeIgniter4", true);
        $this->zrp->traceFunction('CodeIgniter\CodeIgniter::run', [$this, 'beforeRun'], [$this, 'afterRun']);
    }

    /**
     * Before CodeIgniter run, set up trace functions and send template metadata
     */
    public function beforeRun(array $context, array &$storage)
    {
        $this->zrp->traceFunction('CodeIgniter\Debug\Toolbar::run', null, [$this, 'afterToolbarRun']);
        $this->zrp->traceFunction('CodeIgniter\HTTP\Response::send', null, [$this, 'body']);
        $this->zrp->traceFunction('CodeIgniter\Autoloader\FileLocator::locateFile', null, [$this, 'locateHelper']);
        $this->zrp->traceFunction('CodeIgniter\Autoloader\FileLocator::search', null, [$this, 'searchHelper']);
        $this->zrp->traceFunction('CodeIgniter\Events\Events::on', null, [$this, 'afterEventsListener']);
        $this->zrp->traceFunction('CodeIgniter\Events\Events::trigger', null, [$this, 'afterEventTrigger']);

        // load plugin layout data
        if (is_readable(__DIR__ . '/meta.json')) {
            try {
                $meta = json_decode(file_get_contents(__DIR__ . '/meta.json'), true, 512, JSON_THROW_ON_ERROR);
                $storage['plugin'][] = $meta;
            } catch (JsonException $e) {
                // ignore
            }
        }
    }

    /**
     * After CodeIgniter collect Toolbar debug data when CodeIgniter is not in debug mode
     * and send all collected data to ZRay
     */
    public function afterRun(array $context, array &$storage)
    {
        $this->doToolbarRun();

        foreach ($this->collected as $key => $item) {
            if (!empty($item)) {
                $storage[$key][] = $item;
            }
        }
    }

    /**
     * Collect Toolbar debug data in production mode
     */
    private function doToolbarRun(): void
    {
        if (defined('CI_DEBUG') && constant('CI_DEBUG')) {
            //if we are in debug mode and Toolbar is enabled then data collection is done in afterToolbarRun
            return;
        }
        if (
            !class_exists('\Config\Services', false)
            || !class_exists('\Config\Toolbar', false)
            || !function_exists('config')
        ) {
            return;
        }

        $app = \Config\Services::codeigniter();

        $request ??= \Config\Services::request();
        $response ??= \Config\Services::response();

        $toolbar = \Config\Services::toolbar(\config(\Config\Toolbar::class));

        try {
            $stats = $app->getPerformanceStats();
            $dataJson = $toolbar->run(
                $stats['startTime'],
                $stats['totalTime'],
                $request,
                $response
            );
            $data = json_decode($dataJson, true, 512, JSON_THROW_ON_ERROR);
            $this->processToolbarData($data);
        } catch (\Exception $e) {
        }
    }

    /**
     * Collect Toolbar debug data when CodeIgniter is in debug mode
     */
    public function afterToolbarRun(array $context, array &$storage)
    {
        $data = $context['locals']['data'] ?? [];
        $this->processToolbarData($data);
    }

    /**
     * Process collected Toolbar debug data
     */
    private function processToolbarData(array $data)
    {
        if (array_key_exists('collectors', $data) && is_array($data['collectors'])) {
            foreach ($data['collectors'] as $collector) {
                if (!isset($collector['title']) || !isset($collector['display'])) {
                    continue;
                }
                if ($collector['title'] === 'Routes') {
                    $this->collected['Routes'] = $collector['display']['routes'] ?? [];
                    $data['config']['matchedRoute'] = $collector['display']['matchedRoute'][0];
                }
            }
        }
        if (array_key_exists('config', $data) && is_array($data['config'])) {
            if (defined('CI_DEBUG')) {
                $data['config']['debugMode'] = constant('CI_DEBUG');
            }
            $this->collected['State'] = $data['config'];
        }
    }

    public function locateHelper(array $context, array &$storage)
    {
        if (isset($context['functionArgs'][1]) && $context['functionArgs'][1] === 'Helpers') {
            if ($context['returnValue']) {
                $this->collected['Helpers'][] = [
                    'name' => $context['functionArgs'][0],
                    'paths' => $context['returnValue']
                ];
            }
        }
    }

    /**
     * Collect loaded helpers
     */
    public function searchHelper(array $context, array &$storage)
    {
        $path = $context['functionArgs'][0] ?? '';
        if ($path && substr($path, 0, 8) === 'Helpers/') {
            $paths = $context['returnValue'];
            if ($paths) {
                $this->collected['Helpers'][] = [
                    'name' => substr($path, 8),
                    'paths' => count($paths) > 1 ? $paths : $paths[0]
                ];
            }
        }
    }

    /**
     * Collect response body
     */
    public function body(array $context, array &$storage)
    {
        if (!empty($this->collected['Response']['body'])) {
            return;
        }
        $body = strval($context['this']->getBody());
        if ($body) {
            $this->collected['Response'] = ['body' => $body];
        }
    }

    /**
     * Collect CodeIgniter event listeners
     */
    public function afterEventsListener(array $context, array &$storage)
    {

        $name = $context['functionArgs'][0];
        $callback = $context['functionArgs'][1];
        try {
            if ($callback instanceof \Closure) {
                $reflection = new \ReflectionFunction($callback);
                $callback = strval($reflection);
            } elseif (is_object($callback)) {
                $callback = strval(new \ReflectionMethod($callback));
            } else {
                $callback = strval($callback);
            }
        } catch (\ReflectionException $e) {
            $callback = var_export($callback, true);
        }
        $callback = $this->cleanString($callback);

        $this->collected['Listeners'][] = compact('name', 'callback');
    }

    /**
     * Collect events data
     */
    public function afterEventTrigger(array $context, array &$storage)
    {
        $locals = $context['locals'];
        $this->collected['Events'][] = [
            'event' => $locals['eventName'],
            'listeners' => count($locals['listeners']),
            'arguments' => empty($locals['arguments']) ? '-' : json_encode($locals['arguments']),
            'start_time' => empty($locals['start']) ? '' : date('Y-m-d H:i:s', intval($locals['start']))
        ];
    }

    /**
     * Remove new lines and replaces backslashes with an utf8 symbol to avoid showing double backslashes
     */
    private function cleanString(string $string): string
    {
        return str_replace(["\n", "\\"], ['', "⧹"], $string);
    }
}
