<?php

/**
 * Copyright (c) 2024 Zend by Perforce
 */

namespace ZRay;

use JsonException;
use StdClass;
use MongoDB\Client;
use MongoDB\Driver\Monitoring\CommandFailedEvent;
use MongoDB\Driver\Monitoring\CommandStartedEvent;
use MongoDB\Driver\Monitoring\CommandSubscriber;
use MongoDB\Driver\Monitoring\CommandSucceededEvent;
use ZRayPlugin;

class MongoDb implements CommandSubscriber
{
    // MongoDB\Client instance
    protected ?Client $client = null;

    // MongoDB subscriber events
    protected array $events  = [];
    protected array $counters  = [];
    protected array $queries = [];

    const QUERY_KEYS = [
        "ping" => 'PING <db>',
        "find" => 'FIND FROM <db>.<value>',
        "aggregate" => 'AGGREGATE <db>.<value>',
        "insert" => 'INSERT INTO <db>.<value>',
        "update" => 'UPDATE <db>.<value>',
        "delete" => 'DELETE FROM <db>.<value>',
        "projection" => 'PROJECTION <value>',
        "filter" => 'WHERE <value>',
        "pipeline" => 'PIPELINE <value>',
        "documents" => 'VALUES <value>',
        "updates" => '<condition>',
        "deletes" => '<condition>'
    ];

    public function __construct()
    {
        $zrp = new ZRayPlugin("MongoDB", true);

        $zrp->traceFunction('MongoDB\Client::__construct', null, [$this, 'subscribe']);
        $zrp->traceFunction('ZRay\MongoDb::shutdown', null, [$this, 'collectAllData']);
    }

    /**
     * After new MongoDB\Client instance is created, capture it and set up trace functions
     */
    public function subscribe($context, &$storage)
    {
        //version check
        $ver = constant('MONGODB_VERSION');
        if ($this->client || version_compare($ver, '1.15.0', '<')) {
            // Listen only to first connection. TODO?
            // ignore MongoDB versions older than 1.15
            return;
        }
        if (! isset($context['this'])) {
            return;
        }

        $this->client = $context['this'];
        $this->client->getManager()->addSubscriber($this);
        register_shutdown_function([$this, 'shutdown']);

        if (! is_readable(__DIR__ . '/meta.json')) {
            return;
        }

        // load plugin layout data
        try {
            $meta = json_decode(file_get_contents(__DIR__ . '/meta.json'), true, 512, JSON_THROW_ON_ERROR);
            $storage['plugin'][] = $meta;
        } catch (JsonException $e) {
            // ignore
        }
    }

    /**
     * At the end of the request, send all collected data to ZRay
     */
    public function collectAllData(array $context, array &$storage): void
    {
        if (!$this->client) {
            return;
        }

        $info = method_exists($this->client, '__debugInfo') ? $this->client->__debugInfo() : [];

        $info['version'] = constant('MONGODB_VERSION');
        if (isset($info['manager']) && is_object($info['manager'])) {
            $info['manager'] = get_class($info['manager']);
        }
        if (!empty($this->counters)) {
            $info['write operations'] = $this->counters;
        }

        $storage['info'][] = $info;
        $storage['queries'][] = $this->queries;
    }

    public function shutdown()
    {
    }

    /**
     * Parse MongoDB condition object
     */
    private function conditionParse(StdClass $obj): array
    {
        $mql = [];
        foreach (get_object_vars($obj) as $key => $value) {
            if (!$value) {
                continue;
            }

            switch ($key) {
                case 'q':
                    array_push($mql, 'WHERE', json_encode($value));
                    break;
                case 'u':
                    array_push(
                        $mql,
                        ...(
                            is_object($value) && isset($value->{'$set'})
                            ? ['SET', json_encode($value->{'$set'})]
                            : ['REPLACE', json_encode($value)]
                        )
                    );
                    break;
                default:
                    array_push($mql, strtoupper($key), is_scalar($value) ? $value : json_encode($value));
            }
        }

        return $mql;
    }

    /**
     * Callback function for information about started MongoDB command.
     */
    public function commandStarted(CommandStartedEvent $event): void
    {
        $cmd = $event->getCommand();

        $mql = [];
        $params = get_object_vars($cmd);
        $db = $params['$db'] ?? '<unknown>';
        unset($params['$db'], $params['lsid'], $params['cursor']);

        foreach (self::QUERY_KEYS as $key => $template) {
            if (!isset($params[$key])) {
                continue;
            }
            $value = $params[$key];
            unset($params[$key]);
            if (!$value || empty((array)$value)) {
                continue;
            }
            $value = is_scalar($value) ? $value : json_encode($value);

            // template solving
            if ($template === '<condition>') {
                if (isset($cmd->{$key}[0])) {
                    array_push($mql, ...$this->conditionParse($cmd->{$key}[0]));
                }
            } else {
                $mql[] = str_replace(['<db>', '<value>'], [$db, $value], $template);
            }
        }
        if ($params) {
            // add remaining params
            foreach ($params as $key => $value) {
                if (!$value || empty((array)$value)) {
                    continue;
                }
                array_push($mql, strtoupper($key), is_scalar($value) ? $value : json_encode($value));
            }
        }

        $this->events[$event->getCommandName()][$event->getOperationId()] = ['query' => implode(' ', $mql)];
    }

    /**
     * Callback function for information about successful MongoDB command.
     */
    public function commandSucceeded(CommandSucceededEvent $event): void
    {
        $name = $event->getCommandName();
        $id = $event->getOperationId();

        $success = $this->events[$name][$id] ?? false;
        unset($this->events[$name][$id]);

        if (!$success || !is_array($success)) {
            return;
        }

        $reply = $event->getReply();
        $rows = 0;
        if (isset($reply->cursor->firstBatch) && is_array($reply->cursor->firstBatch)) {
            $rows = count($reply->cursor->firstBatch);
        } elseif (isset($reply->n)) {
            $rows = $reply->n;
            $this->counters[$name] = ($this->counters[$name] ?? 0) + $reply->n;
        }
        $this->queries[] = array_merge(
            $success,
            [
                'id' => count($this->queries) + 1,
                'time' => $event->getDurationMicros() / 1000,
                'status' => 'ok',
                'rows' => strval($rows),
            ]
        );
    }

    /**
     * Callback function for information about failed MongoDB command.
     */
    public function commandFailed(CommandFailedEvent $event): void
    {
        $name = $event->getCommandName();
        $id = $event->getOperationId();

        $fail = $this->events[$name][$id];
        unset($this->events[$name][$id]);

        if (!$fail || !is_array($fail)) {
            return;
        }

        $this->queries[] = array_merge(
            $fail,
            [
                'id' => count($this->queries) + 1,
                'time' => $event->getDurationMicros() / 1000,
                'status' => 'fail',
                'rows' => '',
            ]
        );
    }
}
